@@ -19,7 +19,7 @@
 /* VTE accessibility object.  Based heavily on inspection of libzvt's
  * accessibility code. */
 
-#ident "$Id: vteaccess.c,v 1.39 2003/05/19 20:55:10 nalin Exp $"
+#ident "$Id: vteaccess.c,v 1.44 2004/05/01 07:12:51 nalin Exp $"
 
 #include "../config.h"
 
@@ -29,6 +29,7 @@
 #include "debug.h"
 #include "vte.h"
 #include "vteaccess.h"
+#include "vteint.h"
 
 #ifdef HAVE_LOCALE_H
 #include <locale.h>
@@ -61,6 +62,8 @@ enum direction {
 static gunichar vte_terminal_accessible_get_character_at_offset(AtkText *text,
 								gint offset);
 
+static gpointer parent_class = NULL;
+
 /* Create snapshot private data. */
 static VteTerminalAccessiblePrivate *
 vte_terminal_accessible_new_private_data(void)
@@ -101,6 +104,61 @@ vte_terminal_accessible_free_private_data(VteTerminalAccessiblePrivate *priv)
 	g_free(priv);
 }
 
+static gint
+offset_from_xy (VteTerminalAccessiblePrivate *priv,
+		gint x, gint y)
+{
+	gint offset;
+	gint linebreak;
+	gint next_linebreak;
+
+	if (y >= priv->snapshot_linebreaks->len)
+		y = priv->snapshot_linebreaks->len -1;
+
+	linebreak = g_array_index (priv->snapshot_linebreaks, int, y);
+	if (y +1 == priv->snapshot_linebreaks->len)
+		next_linebreak = priv->snapshot_characters->len;
+	else
+		next_linebreak = g_array_index (priv->snapshot_linebreaks, int, y + 1);
+
+	offset = linebreak + x;
+	if (offset >= next_linebreak)
+		offset = next_linebreak -1;
+	return offset;
+}
+
+static void
+xy_from_offset (VteTerminalAccessiblePrivate *priv,
+		gint offset, gint *x, gint *y)
+{
+	gint i;
+	gint linebreak;
+	gint cur_x, cur_y;
+	gint cur_offset = 0;
+
+	cur_x = -1;
+	cur_y = -1;
+	for (i = 0; i < priv->snapshot_linebreaks->len; i++) {
+		linebreak = g_array_index (priv->snapshot_linebreaks, int, i);
+		if (offset < linebreak) {
+			cur_x = offset - cur_offset;
+			cur_y = i - 1;
+			break;
+
+		}  else {
+			cur_offset = linebreak;
+		}
+	}
+	if (i == priv->snapshot_linebreaks->len) {
+		if (offset < priv->snapshot_characters->len) {
+			cur_x = offset - cur_offset;
+			cur_y = i - 1;
+		}
+	}
+	*x = cur_x;
+	*y = cur_y;
+}
+
 /* "Oh yeah, that's selected.  Sure." callback. */
 static gboolean
 all_selected(VteTerminal *terminal, glong column, glong row, gpointer data)
@@ -287,10 +345,10 @@ vte_terminal_accessible_update_private_data_if_needed(AtkObject *text,
 		priv->snapshot_linebreaks = g_array_new(FALSE, TRUE, sizeof(int));
 
 		/* Get a new view of the uber-label. */
-		tmp = vte_terminal_get_text(terminal,
-					    all_selected,
-					    NULL,
-					    priv->snapshot_attributes);
+		tmp = vte_terminal_get_text_include_trailing_spaces(terminal,
+								    all_selected,
+								    NULL,
+								    priv->snapshot_attributes);
 		if (tmp == NULL) {
 			/* Aaargh!  We're screwed. */
 			return;
@@ -676,6 +734,72 @@ vte_terminal_accessible_visibility_notify(VteTerminal *terminal,
 				       visible);
 }
 
+static void
+vte_terminal_initialize (AtkObject *obj, gpointer data)
+{
+	VteTerminal *terminal;
+	AtkObject *parent;
+
+	ATK_OBJECT_CLASS (parent_class)->initialize (obj, data);
+
+	terminal = VTE_TERMINAL (data);
+
+	_vte_terminal_accessible_ref(terminal);
+
+	g_object_set_data(G_OBJECT(obj),
+			  VTE_TERMINAL_ACCESSIBLE_PRIVATE_DATA,
+			  vte_terminal_accessible_new_private_data());
+
+	g_signal_connect(G_OBJECT(terminal), "text-inserted",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_modified),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "text-deleted",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_modified),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "text-modified",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_modified),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "text-scrolled",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_scrolled),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "cursor-moved",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_invalidate_cursor),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "window-title-changed",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_title_changed),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "focus-in-event",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_focus_in),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "focus-out-event",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_focus_out),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "visibility-notify-event",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_visibility_notify),
+			 obj);
+
+	if (GTK_IS_WIDGET((GTK_WIDGET(terminal))->parent)) {
+		parent = gtk_widget_get_accessible((GTK_WIDGET(terminal))->parent);
+		if (ATK_IS_OBJECT(parent)) {
+			atk_object_set_parent(obj, parent);
+		}
+	}
+
+	atk_object_set_name(obj, "Terminal");
+	atk_object_set_description(obj,
+				   terminal->window_title ?
+				   terminal->window_title :
+				   "");
+
+	atk_object_notify_state_change(obj,
+				       ATK_STATE_FOCUSABLE, TRUE);
+	atk_object_notify_state_change(obj,
+				       ATK_STATE_EXPANDABLE, FALSE);
+	atk_object_notify_state_change(obj,
+				       ATK_STATE_RESIZABLE, TRUE);
+	obj->role = ATK_ROLE_TERMINAL;
+}
+
 /**
  * vte_terminal_accessible_new:
  * @terminal: a #VteTerminal
@@ -687,76 +811,16 @@ vte_terminal_accessible_visibility_notify(VteTerminal *terminal,
 AtkObject *
 vte_terminal_accessible_new(VteTerminal *terminal)
 {
-	GtkAccessible *access;
-	AtkObject *parent;
+	AtkObject *accessible;
 	GObject *object;
 
 	g_return_val_if_fail(VTE_IS_TERMINAL(terminal), NULL);
 
 	object = g_object_new(VTE_TYPE_TERMINAL_ACCESSIBLE, NULL);
-	g_return_val_if_fail(GTK_IS_ACCESSIBLE(object), NULL);
+	accessible = ATK_OBJECT (object);
+	atk_object_initialize(accessible, terminal);
 
-	access = GTK_ACCESSIBLE(object);
-	atk_object_initialize(ATK_OBJECT(access), G_OBJECT(terminal));
-
-	access->widget = GTK_WIDGET(terminal);
-	_vte_terminal_accessible_ref(terminal);
-	g_object_add_weak_pointer(G_OBJECT(terminal),
-				  (gpointer*)&access->widget);
-
-	g_object_set_data(G_OBJECT(access),
-			  VTE_TERMINAL_ACCESSIBLE_PRIVATE_DATA,
-			  vte_terminal_accessible_new_private_data());
-
-	g_signal_connect(G_OBJECT(terminal), "text-inserted",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_modified),
-			 object);
-	g_signal_connect(G_OBJECT(terminal), "text-deleted",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_modified),
-			 object);
-	g_signal_connect(G_OBJECT(terminal), "text-modified",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_modified),
-			 object);
-	g_signal_connect(G_OBJECT(terminal), "text-scrolled",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_scrolled),
-			 object);
-	g_signal_connect(G_OBJECT(terminal), "cursor-moved",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_invalidate_cursor),
-			 object);
-	g_signal_connect(G_OBJECT(terminal), "window-title-changed",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_title_changed),
-			 access);
-	g_signal_connect(G_OBJECT(terminal), "focus-in-event",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_focus_in),
-			 access);
-	g_signal_connect(G_OBJECT(terminal), "focus-out-event",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_focus_out),
-			 access);
-	g_signal_connect(G_OBJECT(terminal), "visibility-notify-event",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_visibility_notify),
-			 access);
-
-	if (GTK_IS_WIDGET((GTK_WIDGET(terminal))->parent)) {
-		parent = gtk_widget_get_accessible((GTK_WIDGET(terminal))->parent);
-		if (ATK_IS_OBJECT(parent)) {
-			atk_object_set_parent(ATK_OBJECT(access), parent);
-		}
-	}
-
-	atk_object_set_name(ATK_OBJECT(access), "Terminal");
-	atk_object_set_description(ATK_OBJECT(access),
-				   terminal->window_title ?
-				   terminal->window_title :
-				   "");
-
-	atk_object_notify_state_change(ATK_OBJECT(access),
-				       ATK_STATE_FOCUSABLE, TRUE);
-	atk_object_notify_state_change(ATK_OBJECT(access),
-				       ATK_STATE_EXPANDABLE, FALSE);
-	atk_object_notify_state_change(ATK_OBJECT(access),
-				       ATK_STATE_RESIZABLE, TRUE);
-
-	return ATK_OBJECT(access);
+	return accessible;
 }
 
 static void
@@ -785,19 +849,19 @@ vte_terminal_accessible_finalize(GObject *object)
 						     G_SIGNAL_MATCH_FUNC |
 						     G_SIGNAL_MATCH_DATA,
 						     0, 0, NULL,
-						     vte_terminal_accessible_text_modified,
+						     (gpointer)vte_terminal_accessible_text_modified,
 						     object);
 		g_signal_handlers_disconnect_matched(G_OBJECT(accessible->widget),
 						     G_SIGNAL_MATCH_FUNC |
 						     G_SIGNAL_MATCH_DATA,
 						     0, 0, NULL,
-						     vte_terminal_accessible_text_scrolled,
+						     (gpointer)vte_terminal_accessible_text_scrolled,
 						     object);
 		g_signal_handlers_disconnect_matched(G_OBJECT(accessible->widget),
 						     G_SIGNAL_MATCH_FUNC |
 						     G_SIGNAL_MATCH_DATA,
 						     0, 0, NULL,
-						     vte_terminal_accessible_invalidate_cursor,
+						     (gpointer)vte_terminal_accessible_invalidate_cursor,
 						     object);
 		g_signal_handlers_disconnect_matched(G_OBJECT(accessible->widget),
 						     G_SIGNAL_MATCH_FUNC |
@@ -1155,24 +1219,90 @@ vte_terminal_accessible_get_caret_offset(AtkText *text)
 	return priv->snapshot_caret;
 }
 
+static AtkAttributeSet *
+get_attribute_set (struct _VteCharAttributes attr)
+{
+	AtkAttributeSet *set = NULL;
+	AtkAttribute *at;
+
+	if (attr.underline) {
+		at = g_new (AtkAttribute, 1);
+		at->name = g_strdup ("underline");
+		at->value = g_strdup ("true");
+		set = g_slist_append (set, at);
+	}
+	if (attr.strikethrough) {
+		at = g_new (AtkAttribute, 1);
+		at->name = g_strdup ("strikethrough");
+		at->value = g_strdup ("true");
+		set = g_slist_append (set, at);
+	}
+	at = g_new (AtkAttribute, 1);
+	at->name = g_strdup ("fg-color");
+	at->value = g_strdup_printf ("%u,%u,%u",
+				     attr.fore.red, attr.fore.green, attr.fore.blue);
+	set = g_slist_append (set, at);
+
+	at = g_new (AtkAttribute, 1);
+	at->name = g_strdup ("bg-color");
+	at->value = g_strdup_printf ("%u,%u,%u",
+				     attr.back.red, attr.back.green, attr.back.blue);
+	set = g_slist_append (set, at);
+
+	return set;
+}
+
 static AtkAttributeSet *
 vte_terminal_accessible_get_run_attributes(AtkText *text, gint offset,
 					   gint *start_offset, gint *end_offset)
 {
-	g_return_val_if_fail(VTE_IS_TERMINAL_ACCESSIBLE(text), NULL);
+	VteTerminalAccessiblePrivate *priv;
+	gint i;
+	struct _VteCharAttributes cur_attr;
+	struct _VteCharAttributes attr;
+
 	vte_terminal_accessible_update_private_data_if_needed(ATK_OBJECT(text),
 							      NULL, NULL);
-	/* FIXME */
-	return NULL;
+
+	priv = g_object_get_data(G_OBJECT(text),
+				 VTE_TERMINAL_ACCESSIBLE_PRIVATE_DATA);
+
+	attr = g_array_index (priv->snapshot_attributes,
+			      struct _VteCharAttributes,
+			      offset);
+	*start_offset = 0;
+	for (i = offset - 1; i >= 0; i--) {
+		cur_attr = g_array_index (priv->snapshot_attributes,
+				      struct _VteCharAttributes,
+				      i);
+		if (!gdk_color_equal (&cur_attr.fore, &attr.fore) ||
+		    !gdk_color_equal (&cur_attr.back, &attr.back) ||
+		    cur_attr.underline != attr.underline ||
+		    cur_attr.strikethrough != attr.strikethrough) {
+			*start_offset = i + 1;
+			break;
+		}
+	}
+	*end_offset = priv->snapshot_attributes->len - 1;
+	for (i = offset + 1; i < priv->snapshot_attributes->len; i++) {
+		cur_attr = g_array_index (priv->snapshot_attributes,
+				      struct _VteCharAttributes,
+				      i);
+		if (!gdk_color_equal (&cur_attr.fore, &attr.fore) ||
+		    !gdk_color_equal (&cur_attr.back, &attr.back) ||
+		    cur_attr.underline != attr.underline ||
+		    cur_attr.strikethrough != attr.strikethrough) {
+			*end_offset = i - 1;
+			break;
+		}
+	}
+
+	return get_attribute_set (attr);
 }
 
 static AtkAttributeSet *
 vte_terminal_accessible_get_default_attributes(AtkText *text)
 {
-	g_return_val_if_fail(VTE_IS_TERMINAL_ACCESSIBLE(text), NULL);
-	vte_terminal_accessible_update_private_data_if_needed(ATK_OBJECT(text),
-							      NULL, NULL);
-	/* FIXME */
 	return NULL;
 }
 
@@ -1182,10 +1312,29 @@ vte_terminal_accessible_get_character_extents(AtkText *text, gint offset,
 					      gint *width, gint *height,
 					      AtkCoordType coords)
 {
+	VteTerminalAccessiblePrivate *priv;
+	VteTerminal *terminal;
+	glong char_width, char_height;
+	gint base_x, base_y;
+
 	g_return_if_fail(VTE_IS_TERMINAL_ACCESSIBLE(text));
+
 	vte_terminal_accessible_update_private_data_if_needed(ATK_OBJECT(text),
 							      NULL, NULL);
-	/* FIXME */
+	priv = g_object_get_data(G_OBJECT(text),
+				 VTE_TERMINAL_ACCESSIBLE_PRIVATE_DATA);
+	terminal = VTE_TERMINAL (GTK_ACCESSIBLE (text)->widget);
+
+	atk_component_get_position (ATK_COMPONENT (text), &base_x, &base_y, coords);
+	xy_from_offset (priv, offset, x, y);
+	char_width = vte_terminal_get_char_width (terminal);
+	char_height = vte_terminal_get_char_height (terminal);
+	*x *= char_width;
+	*y *= char_height;
+	*width = char_width;
+	*height = char_height;
+	*x += base_x;
+	*y += base_y;
 }
 
 static gint
@@ -1207,11 +1356,27 @@ vte_terminal_accessible_get_offset_at_point(AtkText *text,
 					    gint x, gint y,
 					    AtkCoordType coords)
 {
-	g_return_val_if_fail(VTE_IS_TERMINAL_ACCESSIBLE(text), 0);
+	VteTerminalAccessiblePrivate *priv;
+	VteTerminal *terminal;
+	glong char_width, char_height;
+	gint base_x, base_y;
+
+	g_return_val_if_fail(VTE_IS_TERMINAL_ACCESSIBLE(text), -1);
+
 	vte_terminal_accessible_update_private_data_if_needed(ATK_OBJECT(text),
 							      NULL, NULL);
-	/* FIXME */
-	return 0;
+	priv = g_object_get_data(G_OBJECT(text),
+				 VTE_TERMINAL_ACCESSIBLE_PRIVATE_DATA);
+	terminal = VTE_TERMINAL (GTK_ACCESSIBLE (text)->widget);
+
+	atk_component_get_position (ATK_COMPONENT (text), &base_x, &base_y, coords);
+	char_width = vte_terminal_get_char_width (terminal);
+	char_height = vte_terminal_get_char_height (terminal);
+	x -= base_x;
+	y -= base_y;
+	x /= char_width;
+	y /= char_height;
+	return offset_from_xy (priv, x, y);
 }
 
 static gint
@@ -1465,7 +1630,7 @@ vte_terminal_accessible_add_focus_handler(AtkComponent *component,
 				  signal_id,
 				  0,
 				  NULL,
-				  handler,
+				  (gpointer)handler,
 				  NULL) != 0) {
 		return 0;
 	}
@@ -1514,21 +1679,17 @@ static void
 vte_terminal_accessible_class_init(gpointer *klass)
 {
 	GObjectClass *gobject_class;
+	AtkObjectClass *class = ATK_OBJECT_CLASS (klass);
 
 	gobject_class = G_OBJECT_CLASS(klass);
 
+	parent_class = g_type_class_peek_parent (klass);
+
+	class->initialize = vte_terminal_initialize;
 	/* Override the finalize method. */
 	gobject_class->finalize = vte_terminal_accessible_finalize;
 }
 
-static void
-vte_terminal_accessible_init(gpointer *instance, gpointer *klass)
-{
-	/* Mark the role this object plays. */
-	g_return_if_fail(ATK_IS_OBJECT(instance));
-	atk_object_set_role(ATK_OBJECT(instance), ATK_ROLE_TERMINAL);
-}
-
 GType
 vte_terminal_accessible_get_type(void)
 {
@@ -1557,9 +1718,9 @@ vte_terminal_accessible_get_type(void)
 		(GClassFinalizeFunc)NULL,
 		(gconstpointer)NULL,
 
-		sizeof(VteTerminalAccessible),
 		0,
-		(GInstanceInitFunc)vte_terminal_accessible_init,
+		0,
+		(GInstanceInitFunc) NULL,
 
 		(GTypeValueTable*)NULL,
 	};
